import typing
import functools
import dataclasses

import sqlalchemy.ext.asyncio
import collections.abc

from hermitage.definition import contracts as hermitage_contracts

AsyncEngineContract = sqlalchemy.ext.asyncio.AsyncEngine
AsyncConnectionContract = sqlalchemy.ext.asyncio.AsyncConnection
StorageBucketCollectionContract = collections.abc.Iterable[sqlalchemy.Table]

Namespace = hermitage_contracts.Namespace
Address = hermitage_contracts.Address
Invoice = hermitage_contracts.Invoice
InvoiceElement = hermitage_contracts.InvoiceElement
Clause = hermitage_contracts.Clause
Item = hermitage_contracts.Item
Row = hermitage_contracts.Row
LogicalOperator = hermitage_contracts.LogicalOperator
View = hermitage_contracts.View
MetaElement = hermitage_contracts.MetaElement
Plugin = hermitage_contracts.Plugin
ReadPlugin = hermitage_contracts.ReadPlugin
WritePlugin = hermitage_contracts.WritePlugin
PluginRegistry = hermitage_contracts.PluginRegistry


class ReadAsyncConnectionContract(AsyncConnectionContract):
    pass


class WriteAsyncConnectionContract(AsyncConnectionContract):
    pass


class ShutdownContext(typing.TypedDict):
    exc_type: type[Exception]
    exc_val: Exception


class Table:
    def __init__(
        self,
        table: sqlalchemy.Table,
        alias: str | None = None
    ):
        self._table = table
        self.alias = alias.replace('.', '__') if alias else None

    def __call__(
        self,
        column_address: Address | None = None,
        column_alias: str | None = None
    ) -> sqlalchemy.Table | sqlalchemy.Column:
        if column_address is None:
            return self._instance

        column = getattr(self._instance.c, str(column_address))
        if column_alias:
            column = column.label(column_alias)

        return column

    @functools.cached_property
    def _instance(self):
        return self._table.alias(self.alias) if self.alias else self._table

    @functools.cached_property
    def _name(self):
        return self._table.name

    def __str__(self):
        return self._table.name


@dataclasses.dataclass
class Link:
    source_table: Table
    target_table: Table
    source_column: Address
    target_column: Address

    def __str__(self):
        return f'{str(self.source_table)}__{str(self.target_table)}'


@dataclasses.dataclass
class M2O(Link):
    pass


@dataclasses.dataclass
class O2M(Link):
    pass


@dataclasses.dataclass
class M2M(Link):
    interim_table: Table
    interim_source_column: Address
    interim_target_column: Address


class ReadClientContract(typing.Protocol):
    async def __call__(self, invoice: Invoice) -> View: ...


class WriteClientContract(typing.Protocol):
    async def __call__(self, invoice: Invoice) -> View | None: ...



import abc
import collections.abc
import operator
import functools
from collections import deque

import sqlalchemy
import hermitage

from .definition import contracts
from .configuration import Schema


class ClauseCompiler(abc.ABC):
    def __init__(
        self,
        schema: Schema,
    ):
        self._schema = schema

    def __call__(self, condition: contracts.Clause) -> sqlalchemy.ClauseElement | None:
        if condition:
            result = self._compile(condition)
            return result

    @abc.abstractmethod
    def _compile(self, clause: contracts.Clause) -> sqlalchemy.ClauseElement: ...


class FilterCompiler(ClauseCompiler):
    @property
    def _operations(self):
        return {
            hermitage.query.EQ: self._simple_clause(operator.eq),
            hermitage.query.NE: self._simple_clause(operator.ne),
            hermitage.query.LE: self._simple_clause(operator.le),
            hermitage.query.LT: self._simple_clause(operator.lt),
            hermitage.query.GE: self._simple_clause(operator.ge),
            hermitage.query.GT: self._simple_clause(operator.gt),
            hermitage.query.IS: lambda v: self._get_column(v).is_(v.operation.value),
            hermitage.query.LIKE: self._like_clause,
            hermitage.query.NOT: self._not_clause,
            hermitage.query.SET: self._set_clause,
            hermitage.query.RANGE: self._range_clause,
        }

    def _compile(self, clause: contracts.Clause) -> sqlalchemy.ClauseElement:
        if clause.address:
            return self._operations[type(clause.operation)](clause)

        stack = deque()
        for element in clause:
            if element is contracts.LogicalOperator.AND:
                a = stack.pop()
                b = stack.pop()
                stack.append(sqlalchemy.and_(a, b))
            elif element is contracts.LogicalOperator.OR:
                stack.append(sqlalchemy.or_(stack.pop(), stack.pop()))
            else:
                stack.append(self._operations[type(element.operation)](element))

        return stack.pop()

    def _get_column(self, clause: contracts.Clause) -> sqlalchemy.Column:
        return self._schema.get_column(clause)

    def _not_clause(self, clause: contracts.Clause):
        if isinstance(clause.operation, hermitage.query.IS):
            return self._get_column(clause).isnot(clause.operation.value)
        elif isinstance(clause.operation, hermitage.query.EQ):
            return operator.ne(self._get_column(clause), clause.operation.value)
        elif isinstance(clause.operation, hermitage.query.LIKE):
            return self._like_clause(
                contracts.Clause(clause.address, clause.operation),
                inversion=True
            )
        elif isinstance(clause.operation, hermitage.query.SET):
            return self._set_clause(
                contracts.Clause(clause.address, clause.operation),
                inversion=True
            )
        else:
            return sqlalchemy.not_(self._compile(clause)),

    def _simple_clause(self, op):
        return lambda v: op(self._get_column(v), v.operation.value)

    def _logic_clause(self, op):
        return lambda v: op(self._compile(u) for u in v)

    def _like_clause(self, clause: contracts.Clause, inversion: bool = False):
        column = self._get_column(clause)
        operation = clause.operation
        value = f'%{operation.value}%'
        if operation.case_sensitive:
            return column.notlike(value) if inversion else column.like(value)
        else:
            return column.notilike(value) if inversion else column.ilike(value)

    def _set_clause(self, clause: contracts.Clause, inversion: bool = False):
        value = list(clause.operation.value)
        if inversion:
            return self._get_column(clause).notin_(value)
        else:
            return self._get_column(clause).in_(value)

    def _range_clause(self, clause: contracts.Clause):
        params = [
            contracts.Clause(clause.address, condition).set_namespace(clause.namespace)
            for condition in clause.operation.value
            if condition is not None
        ]
        if len(params) > 1:
            return self._logic_clause(sqlalchemy.and_)(params)
        elif len(params) == 1:
            return params[0]


class OrderCompiler(ClauseCompiler):
    def __call__(self, condition: contracts.Clause) -> list[sqlalchemy.ClauseElement]:
        result = self._compile(condition)
        return result

    @property
    def _operations(self):
        return {
            hermitage.query.ASC: sqlalchemy.asc,
            hermitage.query.DESC: sqlalchemy.desc
        }

    def _compile(self, clause: contracts.Clause) -> list[sqlalchemy.ClauseElement]:
        stack = deque()
        result = []
        for element in clause or ():
            if element is contracts.LogicalOperator.AND:
                while stack:
                    result.append(stack.pop())
            else:
                result.append(
                    self._operations[type(element.operation)](
                        self._schema.get_column(element)
                    )
                )
        return result


class QueryBuilder:
    def __init__(
        self,
        invoice: contracts.Invoice,
        schema: Schema,
        plugins: collections.abc.Mapping[type[contracts.MetaElement], collections.abc.Callable] | None = None
    ):
        self._schema = schema
        self._plugins = plugins
        self._columns = []
        self._joins = []
        self._clauses = invoice.clauses
        self._filter = FilterCompiler(self._schema)(self._filter_clause)
        self._order = OrderCompiler(self._schema)(self._order_clause)
        self._limit = self._limit_clause.operation.value if self._limit_clause else None
        self._offset = self._offset_clause.operation.value if self._offset_clause else None
        self._parse_invoice(invoice)
        self._apply_plugins(invoice)

    @functools.cached_property
    def columns(self) -> list[sqlalchemy.Column]:
        return self._columns

    @functools.cached_property
    def filter(self) -> sqlalchemy.ClauseElement | None:
        return self._filter

    @functools.cached_property
    def order(self) -> list[sqlalchemy.ClauseElement] | None:
        return self._order

    @functools.cached_property
    def limit(self) -> int | None:
        return self._limit

    @functools.cached_property
    def offset(self) -> int | None:
        return self._offset

    @functools.cached_property
    def query(self):
        query = sqlalchemy.select(*self._columns)
        if self._joins:
            for join in self._joins:
                query = query.join_from(*join, isouter=True)
        if self.filter is not None:
            query = query.where(self.filter)
        for condition in self.order or ():
            query = query.order_by(condition)
        if self.limit:
            query = query.limit(self.limit)
        if self.offset:
            query = query.offset(self.offset)
        return query

    def _parse_invoice(
        self,
        invoice: contracts.Invoice,
        prefix: str | None = None
    ):
        table = self._schema.get_table(invoice.namespace)

        for item in filter(lambda x: x.address is not None, invoice.items):
            self._columns.append(
                table(
                    column_address=item.address,
                    column_alias=f'{prefix}__{item.name}' if prefix else item.name
                )
            )

        _joins_index = set()
        for nested_item in filter(lambda x: x.invoice is not None, invoice.items):
            nested_invoice = nested_item.invoice
            nested_namespace = invoice.namespace + nested_invoice.namespace
            if link := self._schema.get_m2o(nested_namespace):
                self._joins.append((
                    link.source_table(),
                    link.target_table(),
                    link.source_table(link.source_column) == link.target_table(link.target_column)
                ))
                self._parse_invoice(
                    invoice=contracts.Invoice(nested_namespace, *nested_invoice),
                    prefix=nested_item.name
                )
                _joins_index.add(str(link))

        for clause in filter(
            lambda x: hermitage.query.FilterBit in x.kind.__mro__ and len(x.namespace) > 1,
            invoice.clauses
        ):
            if link := self._schema.get_m2o(clause.namespace):
                if str(link) not in _joins_index:
                    self._joins.append((
                        link.source_table(),
                        link.target_table(),
                        link.source_table(link.source_column) == link.target_table(link.target_column)
                    ))

    def _apply_plugins(self, invoice: contracts.Invoice):
        for item in invoice.meta:
            if self._plugins and (plugin := self._plugins.get(type(item))):
                self._columns, self._filter, self._order, self._limit, self._offset = plugin()(
                    item,
                    self._columns,
                    self._filter,
                    self._order,
                    self._limit,
                    self._offset
                )

    @functools.cached_property
    def _filter_clause(self) -> contracts.Clause | None:
        clauses = filter(lambda x: hermitage.query.FilterBit in x.kind.__mro__, self._clauses)
        try:
            result = functools.reduce(operator.and_, clauses)
            return result
        except TypeError:
            return

    @functools.cached_property
    def _limit_clause(self) -> contracts.Clause | None:
        for clause in filter(lambda x: isinstance(x.operation, hermitage.query.Limit), self._clauses):
            return clause

    @functools.cached_property
    def _offset_clause(self) -> contracts.Clause | None:
        for clause in filter(lambda x: isinstance(x.operation, hermitage.query.Offset), self._clauses):
            return clause

    @functools.cached_property
    def _order_clause(self) -> contracts.Clause | None:
        clauses = filter(lambda x: hermitage.query.OrderBit in x.kind.__mro__, self._clauses)
        try:
            return functools.reduce(operator.and_, clauses)
        except TypeError:
            return



from .definition.contracts import (
    ReadAsyncConnectionContract,
    WriteAsyncConnectionContract,
    AsyncEngineContract,
    ReadClientContract,
    WriteClientContract,
    ShutdownContext,
    PluginRegistry,
    M2O,
    M2M,
    O2M,
    Table
)

from .configuration import (
    Schema
)

from .execution.fetching import (
    ReadClient,
)

from .execution.storing import (
    WriteClient
)

from .bootstraping import (
    get_engine,
    get_connection,
    close_connection
)

from .assembling import (
    FilterCompiler,
    OrderCompiler,
    QueryBuilder
)

from . import (
    plugins
)



import typing

from sqlalchemy.ext.asyncio import create_async_engine

from .definition import contracts


def get_engine(dsn: str) -> contracts.AsyncEngineContract:
    return typing.cast(
        contracts.AsyncEngineContract,
        create_async_engine(
            url=dsn,
            echo=True,
        )
    )


async def get_connection(engine: contracts.AsyncEngineContract) -> contracts.AsyncConnectionContract:
    return typing.cast(contracts.AsyncConnectionContract, await engine.connect())


async def close_connection(
    connection: contracts.AsyncConnectionContract,
    context: contracts.ShutdownContext
) -> typing.NoReturn:
    if connection:
        if context.get('exc_type'):
            await connection.rollback()
        else:
            await connection.commit()
        await connection.close()



import collections.abc
import functools
import typing

import hermitage

from ..configuration import Schema
from ..assembling import QueryBuilder
from ..definition import contracts

ResultDataType = collections.abc.Iterable[collections.abc.Mapping]


class ReadClient:
    def __init__(
        self,
        schema: Schema,
        connection: contracts.ReadAsyncConnectionContract,
        plugin_registry: contracts.PluginRegistry | None = None
    ):
        self._schema = schema
        self._connection = connection
        self._plugin_registry = plugin_registry

    async def __call__(
        self,
        invoice: contracts.Invoice,
    ) -> contracts.View:
        data, plugins = await self._launch_invoice(invoice)

        meta = {}
        for plugin in plugins:
            meta |= plugin.get_result()

        return contracts.View(
            data=data,
            meta=meta or None
        )

    def _get_links(self, invoice: contracts.Invoice) -> dict[contracts.Item, contracts.Link]:
        result = {}
        for item in list(filter(lambda x: x.invoice is not None, invoice.items)):
            if link := self._schema.get_o2m(item.invoice.namespace):
                result[item] = link
                invoice += contracts.Item(str(link.source_column), link.source_column)
            elif link := self._schema.get_m2m(item.invoice.namespace):
                result[item] = link
        return result

    async def _launch_invoice(
        self,
        invoice: contracts.Invoice
    ) -> tuple[ResultDataType, list[contracts.Plugin]]:
        links = self._get_links(invoice)

        query_builder = QueryBuilder(
            schema=self._schema,
            invoice=invoice,
            plugins=self._get_plugins()
        )

        result = await self._connection.execute(query_builder.query)
        data = result.mappings().fetchall()

        plugins = []
        mappers = [_nestify, _collapse_none]
        for element in invoice.meta:
            if self._plugin_registry and (plugin_class := self._plugin_registry.get(type(element))):
                plugin = plugin_class(data)
                plugins.append(plugin)
                mappers.extend(plugin.get_mappers())

        data = [
            functools.reduce(lambda v, f: f(v), mappers, dict(row))
            for row in data
        ]

        for item, link in links.items():
            data = await self._attach_o2m(item, link, data)
            data = await self._attach_m2m(item, link, data)

        return data, plugins

    async def _attach_m2m(
        self,
        item: contracts.Item,
        link: contracts.Link,
        data: collections.abc.MutableSequence[collections.abc.MutableMapping]
    ):
        if not isinstance(link, contracts.M2M):
            return data
        nested_invoice = contracts.Invoice(
            str(link.interim_table),
            str(link.interim_source_column),
            contracts.Item(
                item.name,
                contracts.Invoice(
                    self._schema.get_fk_table_alias(link.interim_target_column),
                    *item.invoice
                )
            )
        )
        nested_invoice += contracts.Clause(
            link.interim_source_column,
            hermitage.query.SET(*(r.get(str(link.source_column)) for r in data))
        )
        nested_result = await self._launch_invoice(nested_invoice)
        index = collections.defaultdict(list)
        for row in nested_result[0]:
            index[row[str(link.interim_source_column)]].append(
                {k.replace(f'{item.name}__', ''): v for k, v in row.items() if k.startswith(item.name)}
            )
        for row in data:
            row[item.name] = index.get(row[str(link.source_column)], [])
        return data

    async def _attach_o2m(
        self,
        item: contracts.Item,
        link: contracts.Link,
        data: collections.abc.MutableSequence[collections.abc.MutableMapping]
    ):
        if not isinstance(link, contracts.O2M):
            return data
        original_items = set(
            item.name for item in item.invoice.items if item.address is not None
        )
        nested_invoice = item.invoice
        nested_invoice += contracts.Item(
            str(link.target_column),
            link.target_column
        )
        nested_invoice += contracts.Clause(
            link.target_column,
            hermitage.query.SET(*(r.get(str(link.source_column)) for r in data))
        )
        nested_result = await self._launch_invoice(nested_invoice)
        index = collections.defaultdict(list)
        for row in nested_result[0]:
            index[row[str(link.target_column)]].append(
                {k: v for k, v in row.items() if k in original_items}
            )
        for row in data:
            row[item.name] = index.get(row[str(link.source_column)], [])
        return data

    def _get_plugins(self):
        if self._plugin_registry:
            return {
                k: p.get_query_builder_plugin
                for k, p in self._plugin_registry.items()
                if contracts.ReadPlugin in p.__mro__ and p.get_query_builder_plugin is not None
            }


def _nestify(data: typing.Any):
    if isinstance(data, str):
        return data
    elif isinstance(data, bytearray):
        return data
    elif isinstance(data, bytes):
        return data
    elif isinstance(data, collections.abc.Mapping):
        result = {}
        for k, v in data.items():
            _split_rec(k, _nestify(v), result)
        return result
    elif isinstance(data, collections.abc.Iterable):
        return [_nestify(item) for item in data]
    else:
        return data


def _split_rec(k, v, out):
    k, *rest = k.split('__', 1)
    if rest:
        _split_rec(rest[0], v, out.setdefault(k, {}))
    else:
        out[k] = v


def _collapse_none(data: collections.abc.Mapping):
    result = {}
    for k, v in data.items():
        if isinstance(v, collections.abc.Mapping):
            v = _collapse_none(v) if any(_ is not None for _ in v.values()) else None
        if isinstance(v, list):
            v = [_collapse_none(_) for _ in v]
        result[k] = v
    return result



import abc

from ..execution.storing import WriteExecutor
from ..definition import contracts


class WritePlugin(contracts.WritePlugin, abc.ABC):
    def __init__(
        self,
        executor: WriteExecutor
    ):
        self._executor = executor



class FilterException(Exception):
    def __init__(
        self,
        field_name: str,
        field_value: str,
        message: str
    ):
        self.field_name = field_name
        self.field_value = field_value
        super().__init__(f'{field_name}: {message} with value {field_value}')


class ColumnNotFound(Exception):
    def __init__(
        self,
        table_name: str,
        column_name: str
    ):
        self.table_name = table_name
        self.column_name = column_name
        super().__init__(f'{column_name} not found in table {table_name}')


class RangeException(Exception):
    pass


class TypeMismatch(Exception):
    pass



import collections
import typing

import sqlalchemy

from ..definition import contracts

FIELD_NAME = "total"


class Beacon(contracts.MetaElement):
    pass


def query_builder_plugin(
    element: Beacon,
    columns: collections.abc.MutableSequence[sqlalchemy.Column] | None,
    *args
):
    if isinstance(element, Beacon):
        if columns is None:
            columns = []
        columns.append(sqlalchemy.text(f"count(*) over () as {FIELD_NAME}"))

    return columns, *args


def remove_beacon(data: collections.abc.MutableMapping):
    if FIELD_NAME in data:
        del data[FIELD_NAME]
    return data


class TotalPlugin(contracts.ReadPlugin):
    @classmethod
    def get_beacon(cls) -> type[Beacon]:
        return Beacon

    @classmethod
    def get_query_builder_plugin(cls) -> collections.abc.Callable:
        return query_builder_plugin

    def get_mappers(self) -> collections.abc.Iterable[collections.abc.Callable]:
        return [remove_beacon]

    def get_result(self) -> dict[str, typing.Any] | None:
        if self._data:
            return {FIELD_NAME: self._data[0][FIELD_NAME]}
        else:
            return {FIELD_NAME: 0}



import collections.abc

import sqlalchemy

from hermitage.definition import contracts
from ..configuration import Schema
from ..assembling import FilterCompiler
from ..definition import contracts

SaverDataContract = contracts.Row | contracts.Clause


def invoice_separator(
    invoice: contracts.Invoice
) -> tuple[list[contracts.Row] | None, contracts.Clause | None, list[contracts.MetaElement] | None]:
    data = None
    clause = None
    beacons = None
    for element in invoice:
        if isinstance(element, contracts.Row):
            if data is None:
                data = [element]
            else:
                data.append(element)
        elif isinstance(element, contracts.Clause):
            if clause is None:
                clause = element
            else:
                clause += element
        elif isinstance(element, contracts.MetaElement):
            if beacons is None:
                beacons = []
            beacons.append(element)
    return data, clause, beacons


class WriteExecutor:
    def __init__(
        self,
        schema: Schema,
        connection: contracts.WriteAsyncConnectionContract,
    ):
        self._schema = schema
        self._connection = connection

    async def insert(
        self,
        namespace: contracts.Namespace,
        data: list[contracts.Row]
    ):
        table = self._schema.get_table(namespace)
        await self._connection.execute(
            sqlalchemy.insert(table()).values(
                [_.as_dict() for _ in data]
            )
        )

    async def update(
        self,
        namespace: contracts.Namespace,
        data: contracts.Row,
        clause: contracts.Clause
    ):
        table = self._schema.get_table(namespace)
        query = sqlalchemy.update(
            table()
        ).values(
            data.as_dict()
        ).where(
            FilterCompiler(schema=self._schema)(clause)
        )
        await self._connection.execute(query)

    async def delete(
        self,
        namespace: contracts.Namespace,
        clause: contracts.Clause
    ):
        table = self._schema.get_table(namespace)
        query = sqlalchemy.delete(
            table()
        ).where(
            FilterCompiler(schema=self._schema)(clause)
        )
        await self._connection.execute(query)


class WriteClient:
    def __init__(
        self,
        schema: Schema,
        connection: contracts.WriteAsyncConnectionContract,
        plugin_registry: contracts.PluginRegistry | None = None
    ):
        self._executor = WriteExecutor(
            schema=schema,
            connection=connection
        )
        self._plugin_registry = {}
        for beacon_type in (plugin_registry or ()):
            if plugin_class := plugin_registry.get(beacon_type):
                if contracts.WritePlugin in plugin_class.__mro__:
                    self._plugin_registry[beacon_type] = plugin_class(executor=self._executor)

    async def __call__(
        self,
        invoice: contracts.Invoice
    ):
        data, clause, beacons = invoice_separator(invoice)
        invoice = await self._apply_plugins(beacons, invoice)
        data, clause, beacons = invoice_separator(invoice)
        if data is None and clause is not None:
            await self._executor.delete(invoice.namespace, clause)
        elif data is not None and clause is None:
            await self._executor.insert(invoice.namespace, data)
        elif data is not None and clause is not None:
            for row in data:
                await self._executor.update(invoice.namespace, row, clause)

    async def _apply_plugins(
        self,
        beacons: collections.abc.Iterable[contracts.MetaElement],
        invoice: contracts.Invoice
    ):
        for beacon in beacons or ():
            if plugin := self._plugin_registry.get(type(beacon)):
                invoice = await plugin(invoice)
        return invoice



from ..definition import contracts
from .generic import WritePlugin
from ..execution.storing import invoice_separator


class Beacon(contracts.MetaElement):
    def __init__(self, clause: contracts.Clause):
        super().__init__()
        self.clause = clause


class UpsertPlugin(WritePlugin):
    async def __call__(self, invoice: contracts.Invoice) -> contracts.Invoice:
        data, clause, beacons = invoice_separator(invoice)
        beacon = next(iter(filter(lambda x: isinstance(x, Beacon), beacons)))
        await self._executor.delete(
            invoice.namespace,
            beacon.clause.set_namespace(invoice.namespace)
        )
        return contracts.Invoice(
            invoice.namespace,
            *(element for element in invoice.elements if not isinstance(element, contracts.Clause))
        )

    @classmethod
    def get_beacon(cls) -> type[Beacon]:
        return Beacon



import collections.abc
import typing

import sqlalchemy

from .definition import contracts


class Schema:
    def __init__(
        self,
        tables: collections.abc.Iterable[sqlalchemy.Table],
        fk_postfix: str = '_id'
    ):
        self._instances = {}
        self._tables = {}
        self._links = {}
        self._fk_postfix = fk_postfix
        self._fk = collections.defaultdict(dict)
        self._index_tables(tables)

    def register_link(
        self,
        namespace: contracts.Namespace,
        link: contracts.Link
    ):
        self._links[namespace] = link

    def get_table(self, namespace: contracts.Namespace) -> contracts.Table | None:
        if isinstance(namespace, contracts.Namespace):
            return self._resolve_namespace(namespace)[-1]

    def get_column(self, element: contracts.InvoiceElement) -> sqlalchemy.Column | None:
        table = self.get_table(element.namespace)
        if table is not None:
            return table(element.address)

    def get_fk_table_alias(
        self,
        address: contracts.Address
    ) -> str | None:
        return str(address).replace(self._fk_postfix, '')

    def get_m2o(
        self,
        namespace: contracts.Namespace
    ) -> contracts.M2O | None:
        if link := self._links.get(namespace):
            if isinstance(link, contracts.M2O):
                return link
        else:
            chain = self._resolve_namespace(namespace)
            if len(chain) < 2:
                return
            source_table = chain[-2]
            target_table = chain[-1]
            if source_table and target_table:
                if (f_keys := self._fk.get(str(source_table))) and str(namespace.last()) in f_keys:
                    self._links[namespace] = contracts.M2O(
                        source_table=source_table,
                        target_table=target_table,
                        source_column=contracts.Address(f_keys[str(namespace.last())][0].name),
                        target_column=contracts.Address(f_keys[str(namespace.last())][1].name),
                    )
            return self._links.get(namespace)

    def get_m2m(
        self,
        namespace: contracts.Namespace
    ) -> contracts.M2M | None:
        if link := self._links.get(namespace):
            if isinstance(link, contracts.M2M):
                return link
        else:
            chain = self._resolve_namespace(namespace, True) or ()
            if len(chain) < 2:
                return
            source_table = chain[-2]
            target_table = chain[-1]
            if namespace not in self._links:
                for interim_table_name, _map in self._fk.items():
                    if len(_map) < 2:
                        continue
                    source_column = None
                    target_column = None
                    interim_source_column = None
                    interim_target_column = None
                    for _cols in _map.values():
                        if _cols[1].table.name == str(source_table):
                            source_column = contracts.Address(_cols[1].name)
                            interim_source_column = contracts.Address(_cols[0].name)
                        elif _cols[1].table.name == str(target_table):
                            target_column = contracts.Address(_cols[1].name)
                            interim_target_column = contracts.Address(_cols[0].name)

                    if source_column and target_column:
                        self._links[namespace] = contracts.M2M(
                            source_table=source_table,
                            target_table=target_table,
                            source_column=source_column,
                            target_column=target_column,
                            interim_table=self.get_table(contracts.Namespace(interim_table_name)),
                            interim_source_column=interim_source_column,
                            interim_target_column=interim_target_column
                        )
                        return self._links[namespace]

    def get_o2m(
        self,
        namespace: contracts.Namespace,
    ) -> contracts.O2M | None:
        if link := self._links.get(namespace):
            if isinstance(link, contracts.O2M):
                return link
        else:
            chain = self._resolve_namespace(namespace, True) or ()
            if len(chain) < 2:
                return
            source_table = chain[-2]
            target_table = chain[-1]
            if namespace not in self._links:
                fk_keys = self._fk.get(str(target_table))
                if fk_keys:
                    t = str(namespace[-2])
                    if t not in fk_keys:
                        t = next((k for k, v in fk_keys.items() if v[1].table.name == t), None)
                    if t:
                        self._links[namespace] = contracts.O2M(
                            source_table=source_table,
                            target_table=target_table,
                            source_column=contracts.Address(fk_keys[t][1].name),
                            target_column=contracts.Address(fk_keys[t][0].name),
                        )
            return self._links.get(namespace)

    def _index_tables(
        self,
        tables: collections.abc.Iterable[sqlalchemy.Table]
    ) -> typing.NoReturn:
        for table in tables:
            self._instances[table.name] = table
            for fk in table.foreign_keys:
                self._fk[table.name][fk.parent.name.replace(self._fk_postfix, '')] = (fk.parent, fk.column)

    def _resolve_namespace(
        self,
        namespace: contracts.Namespace,
        strict: bool = False
    ) -> list[contracts.Table] | None:
        parts = []
        previous = None
        for _name in namespace:
            _name = str(_name)
            if _name in self._instances:
                parts.append(_name)
            if strict:
                continue
            if fk := self._fk.get(previous, {}).get(_name):
                previous = fk[1].table.name
                parts.append(previous)
            elif fk := self._fk.get(_name, {}).get(previous):
                previous = fk[1].table.name
                parts.insert(-1, previous)
            previous = _name

        if len(parts) == len(namespace):
            result = []
            for i, table_name in enumerate(parts, 1):
                table_alias = str(namespace.part(i))
                if table_alias not in self._tables:
                    self._tables[table_alias] = contracts.Table(
                        self._instances[table_name],
                        table_alias if table_alias != self._instances[table_name].name else None
                    )
                result.append(self._tables[table_alias])
            return result
